"""
Module which contains mock request, as well as return arguments for the OrodhaUserClient class
"""

MOCK_BULK_REQUEST_ARGS = {
    "target_user": None,
    "pageSize": 10,
    "pageNum": 1,
    "targets": None
}

MOCK_BULK_RESPONSE_ARGS = [{
    "_id": "some_id",
    "keycloak_id": "some_keycloak_id",
    "username": "some_username",
}]

MOCK_TOKEN_EXCHANGE_RESPONSE = {
   "access_token" : "some_token_value",
   "refresh_token" : None,
   "expires_in" : "some_timestamp"
 }



import pytest
import requests_mock
from orodha_user_client import OrodhaUserClient
from tests.fixtures.mock_credentials import MOCK_BASE_URL
from tests.fixtures.mock_request_args import MOCK_BULK_REQUEST_ARGS, MOCK_BULK_RESPONSE_ARGS
from tests.conftest import MockEnvironment
from orodha_user_client.exceptions import UrlNotFound, RequestError


def test_bulk_get(mock_credentials_object, mock_orodha_keycloak):
    with MockEnvironment(ORODHA_USER_SERVICE_BASE_URL=MOCK_BASE_URL):
        user_client = OrodhaUserClient(mock_credentials_object)

    with requests_mock.Mocker() as mock_request:
        mock_request.post(f"{MOCK_BASE_URL}/get-bulk-users", json=MOCK_BULK_RESPONSE_ARGS)

        response = user_client.bulk_get(MOCK_BULK_REQUEST_ARGS)
        assert response == MOCK_BULK_RESPONSE_ARGS


def test_user_client_bad_url(mock_credentials_object, mock_orodha_keycloak):
    with MockEnvironment(ORODHA_USER_SERVICE_BASE_URL=""):
        with pytest.raises(UrlNotFound):
            OrodhaUserClient(mock_credentials_object)


def test_bulk_get_request_error(mock_credentials_object, mock_orodha_keycloak):
    with MockEnvironment(ORODHA_USER_SERVICE_BASE_URL=MOCK_BASE_URL):
        user_client = OrodhaUserClient(mock_credentials_object)

    with requests_mock.Mocker() as mock_request:
        mock_request.register_uri(
            'POST',
            f"{MOCK_BASE_URL}/get-bulk-users",
            text='Not Found',
            status_code=404
        )
        with pytest.raises(RequestError):
            user_client.bulk_get(MOCK_BULK_REQUEST_ARGS)



"""
Module containing custom exceptions for the orodha_user_client package.
"""
from http import HTTPStatus

class UrlNotFound(Exception):
    """
    Exception that is thrown when a specified URL
    cannot be found in the environment.
    """
    def __init__(self, message: str = "Service url could not be found in environment"):
        self.message = message
        super().__init__(self.message)


class UnexpectedRequestType(Exception):
    """
    Exception for when the request type given to the request factory does not
    match the list of expected requests types, i.e. PUT, POST, GET, DELETE.
    """
    def __init__(self, message: str = None):
        self.message = message
        super().__init__(self.message)


class RequestError(Exception):
    """
    A general exception for when a request to the user service
    fails for any given reason
    """
    def __init__(
        self,
        message: str=None,
        status_code: HTTPStatus=HTTPStatus.BAD_REQUEST):
        self.message = message
        self.status_code = status_code
        super().__init__(f"{self.message}: {self.status_code}")



"""
Module which contains pytest fixtures for use in testing the orodha_user_client.
"""
import os
from copy import deepcopy
import pytest
from orodha_keycloak import OrodhaCredentials
from tests.fixtures.mock_credentials import MOCK_CREDENTIALS
from tests.fixtures.mock_request_args import MOCK_TOKEN_EXCHANGE_RESPONSE


class MockEnvironment:
    """
    Context manager which is used to make changes to environment
    variables in order to test if our code reacts correctly to different variables
    being present.
    """

    def __init__(self, **kwargs):
        self.old_env = deepcopy(os.environ)
        self.new_env = deepcopy(os.environ)
        self.new_env.update(kwargs)

    def __enter__(self):
        for key, value in self.new_env.items():
            os.environ[key.upper()] = value

    def __exit__(
        self,
        exception_type,
        exception_value,
        exception_traceback
    ):
        os.environ = self.old_env


class MockOrodhaKeycloakClient:
    """Mock OrodhaKeycloakConnection object to return keycloak fixture functions in testing."""

    def __init__(self, *args, **kwargs):
        self.mock_args = args
        self.mock_kwargs = kwargs

    def exchange_token(self, *args, **kwargs):
        return MOCK_TOKEN_EXCHANGE_RESPONSE

@pytest.fixture
def mock_credentials_object():
    credential_object = OrodhaCredentials(**MOCK_CREDENTIALS)
    return credential_object


@pytest.fixture
def mock_orodha_keycloak(mocker):
    """
    Fixture which patches the OrodhaKeycloakClient class for use in testing
    the user client.
    """
    mocker.patch(
        "orodha_keycloak.OrodhaKeycloakClient",
        return_value=MockOrodhaKeycloakClient(),
    )



"""
The main module of the Orodha User Client. Contains UserClient class
which is used to connect to and communicate with the Orodha User Service on behalf
of other Orodha services.
"""
import os
from http import HTTPStatus
import requests
from orodha_user_client.exceptions import (
    UrlNotFound,
    RequestError,
    UnexpectedRequestType
)
import orodha_keycloak


BULK_USER_URL = "get-bulk-users"

def _request_factory(request_type: str):
    """
    Factory function which accepts a request_type string, and
    returns the correct method from the requests module.

    Args:
        request_type(str): The string representation of the type of HTTP request
            we want to make.

    Raises:
        SomeError: If the request_type is not PUT, POST, GET, or DELETE.

    Returns:
        request_method: The method from the requests library that matches the request_type
            argument.
    """
    if not hasattr(requests, request_type.lower()):
        raise UnexpectedRequestType(
            message=f"{request_type.lower()} is not an accepted request type."
        )

    return getattr(requests, request_type.lower())

class OrodhaUserClient:
    """
    The main class for the orodha_user_client package. Allows services
    to interact with the Orodha User Service programmatically.
    """
    def __init__(self, credentials: orodha_keycloak.OrodhaCredentials):
        self.credentials = credentials
        self.keycloak_client = orodha_keycloak.OrodhaKeycloakClient(credentials_object=self.credentials)
        self.base_url = self._get_base_url()


    def bulk_get(self, request_args: dict):
        """
        Function that calls the user service get-bulk-users route on
        behalf of other services in the Orodha namespace.

        Args:
            request_args(dict): A dictionary of request arguments to be
                packaged and sent to the Orodha user service.

        Returns:
            response(dict): A dictionary containing our new token info.
        """
        target_user = request_args.get("target_user")
        if not target_user:
            target_user = self.credentials.client_id

        token_data = self.keycloak_client.exchange_token(target_user=target_user)

        body = {
            "pageSize": request_args.get("pageSize"),
            "pageNum": request_args.get("pageNum"),
            "targets": request_args.get("targets")
        }
        headers = {
            "Content-Type": "application/json",
            "Authorization" : f"Bearer {token_data.get('access_token')}"
        }
        return_value = self._make_raw_request(
            BULK_USER_URL,
            "post",
            body=body,
            headers=headers
        )

        return return_value


    def _get_base_url(self):
        base_url = os.environ.get("ORODHA_USER_SERVICE_BASE_URL")
        if not base_url:
            raise UrlNotFound(
                message="\"ORODHA_USER_SERVICE_BASE_URL\" must be present in environment"
            )
        return base_url.rstrip("/")


    def _make_raw_request(
            self,
            route: str,
            request_type: str,
            **request_args
        ):
        desired_request = _request_factory(request_type)

        response = desired_request(
            f"{self.base_url}/{route.lower()}",
            headers=request_args.get("headers"),
            data=request_args.get("body")
        )

        if response.status_code != HTTPStatus.OK:
            raise RequestError(
                message="A problem was encountered during execution.",
                status_code=response.status_code
            )
        try:
            return_value = response.json()
        except ValueError:
            return_value = response.content
        return return_value



"""Module which contains fixture data for use in testing the orodha_user_client."""

MOCK_CREDENTIALS = {
    "server_url": "http://some_server_url",
    "realm_name": "some_realm_name",
    "client_id": "some_client_id",
    "client_secret_key": "some_secret_key",
    "username": "some_username",
    "password": "some_password",
    }

MOCK_BASE_URL = "http://some_other_url"
